#ifndef bqtStmtHH
#define bqtStmtHH

#include <list>

#include "autoptr"
#include "tokens.hh"
#include "expr.hh"

typedef autoptr<class stmt> stmtp;

class stmt: public ptrable
{
    stmtp prev;
public:
    stmt();
    virtual ~stmt();
    
    void set_prev(stmtp p) { prev=p; }
    const stmtp& get_prev() const { return prev; }
    virtual void Compile(const Scope& scope);
};

class stmt_if: public stmt
{
public:
    exprp condition;
    stmtp thenp;
    std::list<
        std::pair<exprp, stmtp>
             > elseifp;
    stmtp elsep;
public:
    virtual void Compile(const Scope& scope);
};
class stmt_select: public stmt
{
public:
    exprp ref;
    std::list<
        std::pair<exprp, stmtp>
             > cases;
public:
    virtual void Compile(const Scope& scope);
};
class stmt_doloop: public stmt
{
public:
    exprp precondition;
    exprp postcondition;
    stmtp bodyp;
public:
    virtual void Compile(const Scope& scope);
};
class stmt_for: public stmt
{
public:
    exprp var;
    exprp begin, end, step;
    stmtp bodyp;
public:
    virtual void Compile(const Scope& scope);
};
class stmt_return: public stmt
{
public:
    QBtoken token;
    exprp label; // In case of RESUME or RETURN
    // return type:
    // RETURN
    // EXITDEF, EXITDO, EXITSUB, EXITFUNCTION, EXITFOR
    // RESUME, RESUMENEXT
    // STOP, END
public:
    stmt_return(QBtoken tok, exprp l=0) : token(tok), label(l) { }
public:
    virtual void Compile(const Scope& scope);
};
class stmt_goto: public stmt
{
public:
    exprp label;
    stmt_goto(exprp l) : label(l) { }
public:
    virtual void Compile(const Scope& scope);
};
class stmt_gosub: public stmt
{
    exprp label;
public:
    stmt_gosub(exprp l) : label(l) { }
public:
    virtual void Compile(const Scope& scope);
};
class stmt_restore: public stmt
{
    long double label;
public:
    stmt_restore(long double l): label(l) { }
public:
    virtual void Compile(const Scope& scope);
};
class stmt_read: public stmt
{
    exprp target;
public:
    stmt_read(exprp t) : target(t) { }
public:
    virtual void Compile(const Scope& scope);
};
class stmt_let: public stmt
{
    exprp target;
    exprp value;
public:
    stmt_let(exprp t, exprp v): target(t), value(v) { }
public:
    virtual void Compile(const Scope& scope);
};

class stmt_print: public stmt // PRINT, LPRINT, PRINT #, WRITE #, PRINT USING.
{
public:
    exprp target;
    bool is_lprint; // Printer?
    exprp using_string;
    
    std::list<std::pair<QBtoken, exprp> > params;
    
    // Note: When parsing the USING format, SPC/TAB/CRLF do not consume a pattern.
public:
    stmt_print() : is_lprint(false) { }
    
    void SetFile(exprp p)       { target = p; }
    void SetLPRINT()            { is_lprint = true; }
    void SetFormat(exprp p)     { using_string = p; }
    void AppendExpr(exprp p)    { params.push_back(std::make_pair((QBtoken)0, p)); }

    // Cursor position control:
    void AppendSPCfunc(exprp p) { params.push_back(std::make_pair(SPC, p)); }
    void AppendTABfunc(exprp p) { params.push_back(std::make_pair(TAB, p)); }
    void AppendTab()     { params.push_back(std::make_pair(TAB,  (expr*)0)); } // The ',' operator
    void AppendNewline() { params.push_back(std::make_pair(CRLF, (expr*)0)); }
    // For number printing:
    void AppendSpace()   { AppendExpr(new expr_string(" ")); }
    // For WRITE:
    void AppendQuoteCh() { AppendExpr(new expr_string("\"")); }
    void AppendCommaCh() { AppendExpr(new expr_string(",")); }
public:
    virtual void Compile(const Scope& scope);
};

class stmt_input: public stmt // INPUT, INPUT #, LINE INPUT, LINE INPUT #
{
public:
    exprp source;
    bool is_linput; // Line input? Line input has no prompt.
    exprp prompt;   // Default prompt is "?". Not used for LINEINPUT.
    bool keep_cursor;
    std::list<exprp> params;
public:
    stmt_input() : is_linput(false), keep_cursor(false) { }
    
    void SetFile(exprp p)   { source = p; }
    void SetLINEINPUT()     { is_linput = true; }
    void SetPrompt(exprp p) { prompt = p; }
    void SetKeepCursor()    { keep_cursor=true; }
    void AppendTarget(exprp p) { params.push_back(p); }
public:
    virtual void Compile(const Scope& scope);
};

class stmt_fileopen: public stmt
{
    exprp filename;
    QBtoken mode;
    exprp mode_str;
    int access;
    int locking;
    exprp reclen;
public:
    stmt_fileopen(exprp fn, exprp mstr, int ac, int lk, exprp rl) 
        : filename(fn), mode_str(mstr), access(ac), locking(lk), reclen(rl)
    {
    }
    stmt_fileopen(exprp fn, QBtoken md, int ac, int lk, exprp rl) 
        : filename(fn), mode(md), access(ac), locking(lk), reclen(rl)
    {
    }
public:
    virtual void Compile(const Scope& scope);
};

class stmt_redim: public stmt
{
    exprp target;
    std::vector<std::pair<exprp, exprp> > dimensions;
public:
    stmt_redim(exprp t, const std::vector<std::pair<exprp, exprp> >& d)
        : target(t), dimensions(d)
    {
    }
public:
    virtual void Compile(const Scope& scope);
};

class stmt_subcall: public stmt
{
    std::string name;
    std::vector<exprp> params;
public:
    stmt_subcall(const std::string& n) : name(n) { }
    stmt_subcall(const std::string& n,
                 const std::vector<exprp>& p) : name(n), params(p) { }
public:
    virtual void Compile(const Scope& scope);
};

class stmt_label: public stmt
{
    long double label;
public:
    stmt_label() : label(0) { }
    stmt_label(long double f) : label(f) { }
    stmt_label(long long int i) : label(i) { }

public:
    virtual void Compile(const Scope& scope);
};

#endif
